import logging
from logging import DEBUG, INFO, WARNING, ERROR, CRITICAL, FATAL, NOTSET

def set_logging_level(level: str='ERROR'):
    """Set the logging level based on the provided string.

    :param str level: The logging level as a string. Options are 'CRITICAL', 'FATAL', 'ERROR', 'WARNING', 'WARN', 'INFO', 'DEBUG', or 'NOTSET'.
    """
    level = level.upper()
    
    logging.basicConfig(level=eval(level))

    logging.info(f"set logging level = {level}")


import logging
import multiprocessing
from ._count import compute_ref_point_matrix, compute_scale_region_matrix

def multi_region_compute_ref_point_matrix(
    input_bed_list: list,
    input_bam: str,
    ref_point: str,
    up_extend_length: int,
    down_extend_length: int,
    extend_binsize: int,
    count_norm_method: str,
    processes: int,
) -> list:
    """Compute count matrix of reference point.

    Args:
        input_bed_list (list): a list of temp bed files.
        input_bam (str): input bam file
        ref_point (str): reference point, TSS, TES, or center.
        up_extend_length (int): upstream region extend length.
        down_extend_length (int): downstream region extend length.
        extend_binsize (int): region extend binsize.
        count_norm_method (str): Can be RPKM, Raw or CPM.
        processes (int): Number of processes to use.

    Returns:
        list: a list containing region information and signal values of each region.
    """
    # Placeholder for actual implementation
    logging.info(f"Counting matrices in {input_bam} with {processes} threads.")

    pool = multiprocessing.Pool(processes=processes)

    # record info
    process_info_list = []

    # count reads for each region
    for input_bed in input_bed_list:
        process_info = pool.apply_async(
            func=compute_ref_point_matrix,
            args=(
                input_bed,
                input_bam,
                ref_point,
                up_extend_length,
                down_extend_length,
                extend_binsize,
                count_norm_method,
            ),
        )
        process_info_list.append(process_info)

    # wait for all tasks to complete
    pool.close()
    pool.join()

    # out pool result
    merge_process_info_list = []
    for process_info in process_info_list:
        merge_process_info_list.extend(process_info.get())
    merge_process_info_list.sort(key=lambda x: sum(x[6:] ) / (len(x) - 6), reverse=True)
    logging.info(f"{input_bam} calculation done!")

    return merge_process_info_list

def multi_region_compute_scale_region_matrix(
    input_bed_list: list,
    input_bam: str,
    up_extend_length: int,
    down_extend_length: int,
    extend_binsize: int,
    split_num: int,
    count_norm_method: str,
    processes: int,
) -> list:
    """Compute count matrix of reference point.

    Args:
        input_bed_list (list): a list of temp bed files.
        input_bam (str): input bam file
        up_extend_length (int): upstream region extend length.
        down_extend_length (int): downstream region extend length.
        extend_binsize (int): region extend binsize.
        split_num (int): number of bins to split region into.
        count_norm_method (str): Can be RPKM, Raw or CPM.
        processes (int): Number of processes to use.

    Returns:
        list: a list containing region information and signal values of each region.
    """
    # Placeholder for actual implementation
    logging.info(f"Counting matrices in {input_bam} with {processes} threads.")

    pool = multiprocessing.Pool(processes=processes)

    # record info
    process_info_list = []

    # count reads for each region
    for input_bed in input_bed_list:
        process_info = pool.apply_async(
            func=compute_scale_region_matrix,
            args=(
                input_bed,
                input_bam,
                up_extend_length,
                down_extend_length,
                extend_binsize,
                split_num,
                count_norm_method,
            ),
        )
        process_info_list.append(process_info)

    # wait for all tasks to complete
    pool.close()
    pool.join()

    # out pool result
    merge_process_info_list = []
    for process_info in process_info_list:
        merge_process_info_list.extend(process_info.get())
    merge_process_info_list.sort(key=lambda x: sum(x[6:] ) / (len(x) - 6), reverse=True)
    logging.info(f"{input_bam} calculation done!")

    return merge_process_info_list


import os
import logging
import random
import gzip
import string

def split_file_and_make_temp(input_filename: str, temp_dir: str = None, n_part: int = 1) -> list:
    """
    Split input file into n_part temporary files.
    
    :param input_filename: Input bed file to be split.
    :param temp_dir: Directory to store temporary files. If None, uses the same directory as input file.
    :param n_part: Number of parts to split the input file into.
    :return: List of paths to the temporary files created.
    """
    if temp_dir is None:
        temp_dir = os.path.abspath(os.path.dirname(input_filename))
    else:
        temp_dir = os.path.abspath(temp_dir)

    if not os.path.exists(temp_dir):
        os.makedirs(temp_dir)
        logging.info(f"Created temporary directory: {temp_dir}")

    if temp_dir[-1] != "/":
        temp_dir += "/"

    # count the number of lines in the input file
    logging.info(f"Counting lines in input file: {input_filename}")
    input = open(input_filename, 'rt') if not input_filename.endswith('.gz') else gzip.open(input_filename, 'rt')

    line_count = 0
    for line in input:
        line_count += 1

    input.close()

    if line_count % n_part == 0:
        lines_per_file = line_count // n_part
    else:
        lines_per_file = line_count // n_part + 1

    # create temporary files
    logging.info(f"Splitting input file into {n_part} parts, each with approximately {lines_per_file} lines.")

    input = open(input_filename, 'rt') if not input_filename.endswith('.gz') else gzip.open(input_filename, 'rt')
    input_file_basename = os.path.basename(input_filename)

    temp_filename_list = []
    temp_file_list = []

    for i in range(n_part):
        temp_file_basename = "temp_" + input_file_basename + "." + str(i) + "." + "".join(random.sample(string.ascii_letters + string.digits, 16))
        temp_file_name = os.path.join(temp_dir, temp_file_basename)
        temp_filename_list.append(temp_file_name)

        temp_file = open(temp_file_name, 'wt') if not input_filename.endswith('.gz') else gzip.open(temp_file_name, 'wt')
        temp_file_list.append(temp_file)

    # write lines to temporary files
    for index, line in enumerate(input):
        file_index = index // lines_per_file
        temp_file_list[file_index].write(line)

    logging.info(f"Finished writing lines to temporary files.")

    input.close()
    [file.close() for file in temp_file_list]

    return temp_filename_list


import seaborn as sns
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt

def plot_heatmap_and_profile(
    matrix_file: str,
    output_heatmap: str,
    output_profile: str,
    vmin: int = 0,
    vmax: int = None,
    cmap: str = "Purples",
):
    """Plot heatmap and profile from csv file.

    Args:
        matrix_file (str): csv file containing the matrix data.
        output_heatmap (str): Output heatmap pdf file.
        output_profile (str): Output profile pdf file.
        vmin (int, optional): Lower boundary of scale bar. Defaults to 0.
        vmax (int, optional): Upper boundary of scale bar. Defaults to None.
        cmap (str, optional): Color theme of heatmap. Defaults to "Purples".
    """
    df = pd.read_csv(matrix_file)

    # heatmap plot
    plt.rcParams["figure.figsize"] = [5, 10]
    plt.rcParams["figure.autolayout"] = True

    if not vmax:
        vmax = np.quantile(df.iloc[:, 6:].values, 0.99)
    p1 = sns.heatmap(df.iloc[:, 6:], vmin=vmin, vmax=vmax, cmap=cmap, xticklabels=False, yticklabels=False)
    fig_p1 = p1.get_figure()
    fig_p1.savefig(output_heatmap, dpi=300)
    fig_p1.clf()

    a = df.iloc[:, 6:].sum(0) / 61

    # profile plot
    plt.rcParams["figure.figsize"] = [5, 5]
    plt.rcParams["figure.autolayout"] = True

    p2 = sns.lineplot(a)
    p2.set_xticklabels([])
    fig_p2 = p2.get_figure()
    fig_p2.savefig(output_profile, dpi=300)
    fig_p2.clf()


import gzip
import os
import sys
import logging

from ._logging import set_logging_level
from ._split import split_file_and_make_temp
from ._multiprocess import multi_region_compute_ref_point_matrix
from ._error import RemoveTempError
from ._plot import plot_heatmap_and_profile

def main(
    input: str,
    input_bam: str,
    output: str,
    output_heatmap: str,
    output_profile: str,
    ref_point: str = "TSS",
    up_extend_length: int = 3000,
    down_extend_length: int = 3000,
    extend_binsize: int = 100,
    count_norm_method: str = "RPKM",
    verbose: str = "ERROR",
    processes: int = 1,
    temp_dir: str = None,
    vmin: int = 0,
    vmax: int = None,
    cmap: str = "Purples",
):
    # set up logging level
    set_logging_level(level=verbose)

    region_split_filename_list = split_file_and_make_temp(
        input_filename=input, temp_dir=temp_dir, n_part=processes
    )

    all_info_list = multi_region_compute_ref_point_matrix(
        input_bed_list=region_split_filename_list,
        input_bam=input_bam,
        ref_point=ref_point,
        up_extend_length=up_extend_length,
        down_extend_length=down_extend_length,
        extend_binsize=extend_binsize,
        count_norm_method=count_norm_method,
        processes=processes,
    )

    # merge file
    out_file = open(output, "wt") if not output.endswith(".gz") else gzip.open(output, "wt")
    bed_header = ["chrom", "start", "end", "name", "score", "strand"]
    scale_region_headers = [f"bin_{i+1}" for i in range(len(all_info_list[0]) - 6)]
    out_header = bed_header + scale_region_headers
    out_header_str = ",".join(out_header)
    out_file.write(out_header_str + "\n")

    for info in all_info_list:
        out_line = ",".join(map(str, info))
        out_file.write(out_line + "\n")

    # ---------------------------->>>
    # close and clean temp file
    # ---------------------------->>>
    sys.stderr.write("-" * 80 + "\n")
    logging.info("Removing temp files...")

    for temp_filename in region_split_filename_list:
        if os.path.exists(temp_filename):
            if os.path.isfile(temp_filename):
                try:
                    os.remove(temp_filename)
                except Exception:
                    raise RemoveTempError

    out_file.close()

    # plot heatmap and profile
    plot_heatmap_and_profile(
        matrix_file=output,
        output_heatmap=output_heatmap,
        output_profile=output_profile,
        vmin=vmin,
        vmax=vmax,
        cmap=cmap,
    )

    logging.info("Everything is done!")


import pysam

from ._error import ParsingError

def query_region_count(
    bam_obj: pysam.AlignmentFile,
    query_region_list: list,
    norm_method: str="CPM",
) -> list:
    """Calculate the normalized read count for given query regions.

    Args:
        bam_obj (pysam.AlignmentFile): _pysam_ AlignmentFile object representing the BAM file.
        query_region_list (list): List of query regions, each defined as a tuple (chromosome, start, end).
        norm_method (str, optional): "CPM", "FPKM" or "Raw". Defaults to "CPM".
    Returns:
        list: List of normalized counts for each query region.
    """
    # count total 
    total_count = 0

    for info in bam_obj.get_index_statistics():
        total_count += info.mapped

    scale_value_list = []
    for region in query_region_list:
        
        region_chr_name = region[0]
        region_start = int(region[1])
        region_end = int(region[2])
        
        region_count = 0
        for align in bam_obj.fetch(contig=region_chr_name, start=region_start, end=region_end):
            region_count += 1
        
        if norm_method == "Raw":
            scale_value = region_count
        elif norm_method == "RPKM":
            region_length = region_end - region_start
            scale_value = region_count / 1.0 / (total_count / 1e6) / (region_length / 1e3)
        elif norm_method == "CPM":
            scale_value = region_count / 1.0 / (total_count / 1e6)
        else:
            raise ValueError("norm_method should be 'Raw', 'RPKM' or 'CPM'")

        scale_value_list.append(round(scale_value, 6))

    return scale_value_list


def compute_ref_point_matrix(
    input_bed: str,
    input_bam: str,
    ref_point: str,
    up_extend_length: int,
    down_extend_length: int,
    extend_binsize: int,
    count_norm_method: str,
) -> list:
    """Compute count matrix of reference point for a single region file.

    Args:
        input_bed (str): a temp bed file.
        input_bam (str): input bam file
        ref_point (str): reference point, TSS, TES, or center.
        up_extend_length (int): upstream region extend length.
        down_extend_length (int): downstream region extend length.
        extend_binsize (int): region extend binsize.
        count_norm_method (str): Can be RPKM, Raw or CPM.

    Returns:
        list: a list containing region information and signal values of each region.
    """
    in_bed_file = open(input_bed, "rt")
    bam_file = pysam.AlignmentFile(input_bam, "rb")

    all_info_list = []

    for region in in_bed_file:
        region_list = region.rstrip().split("\t")
        chrom = region_list[0]
        strand = region_list[5]

        if ref_point == "TSS":
            reference_point = int(region_list[1]) if strand == "+" else int(region_list[2])
        elif ref_point == "TES":
            reference_point = int(region_list[2]) if strand == "+" else int(region_list[1])
        elif ref_point == "center":
            reference_point = int((int(region_list[1]) + int(region_list[2])) / 2)
        else:
            raise ParsingError("ref_point should be TSS, TES, or center.")

        upstream_region_list = []
        downstream_region_list = []
        query_region_list = []

        if strand == "-":
            up_extend_length, down_extend_length = down_extend_length, up_extend_length

        # upstream
        for bin_start in range(reference_point - up_extend_length, reference_point, extend_binsize):
            bin_end = min(bin_start + extend_binsize, reference_point)
            upstream_region_list.append((chrom, bin_start, bin_end))

        # downstream
        for bin_start in range(reference_point, reference_point + down_extend_length, extend_binsize):
            bin_end = min(bin_start + extend_binsize, reference_point + down_extend_length)
            downstream_region_list.append((chrom, bin_start, bin_end))

        # reference point bin
        query_region_list.append((chrom, reference_point - extend_binsize / 2, reference_point + extend_binsize / 2))

        up_signal_val_list = query_region_count(
            bam_obj=bam_file,
            query_region_list=upstream_region_list,
            norm_method=count_norm_method,
        )

        query_signal_val_list = query_region_count(
            bam_obj=bam_file,
            query_region_list=query_region_list,
            norm_method=count_norm_method,
        )

        down_signal_val_list = query_region_count(
            bam_obj=bam_file,
            query_region_list=downstream_region_list,
            norm_method=count_norm_method,
        )

        merge_sinal_val_list = up_signal_val_list + query_signal_val_list + down_signal_val_list
        if strand == "-":
            merge_sinal_val_list.reverse()
        
        all_info_list.append(region_list + merge_sinal_val_list)


    in_bed_file.close()
    bam_file.close()
    return all_info_list


def compute_scale_region_matrix(
    input_bed: str,
    input_bam: str,
    up_extend_length: int,
    down_extend_length: int,
    extend_binsize: int,
    split_num: int,
    count_norm_method: str,
) -> list:
    """Compute count matrix of reference point for a single region file.

    Args:
        input_bed (str): a temp bed file.
        input_bam (str): input bam file
        up_extend_length (int): upstream region extend length.
        down_extend_length (int): downstream region extend length.
        extend_binsize (int): region extend binsize.
        split_num (int): number of bins to split region into.
        count_norm_method (str): Can be RPKM, Raw or CPM.

    Returns:
        list: a list containing region information and signal values of each region.
    """
    in_bed_file = open(input_bed, "rt")
    bam_file = pysam.AlignmentFile(input_bam, "rb")

    all_info_list = []

    for region in in_bed_file:
        region_list = region.rstrip().split("\t")
        chrom = region_list[0]
        strand = region_list[5]
        start = int(region_list[1])
        end = int(region_list[2])

        upstream_region_list = []
        downstream_region_list = []
        query_region_list = []

        if strand == "-":
            up_extend_length, down_extend_length = down_extend_length, up_extend_length

        # upstream
        for bin_start in range(start - up_extend_length, start, extend_binsize):
            bin_end = min(bin_start + extend_binsize, start)
            upstream_region_list.append((chrom, bin_start, bin_end))

        # downstream
        for bin_start in range(end, end + down_extend_length, extend_binsize):
            bin_end = min(bin_start + extend_binsize, end + down_extend_length)
            downstream_region_list.append((chrom, bin_start, bin_end))

        # gene body
        gene_body_length = end - start + 1
        bin_size = gene_body_length / split_num

        bin_start = start
        bin_count = 0
        while bin_start < end:
            bin_end = min(int(bin_start + bin_size), end)
            query_region_list.append((chrom, bin_start, bin_end))
            bin_start = bin_end
            bin_count += 1
            if bin_count >= split_num:
                break

        up_signal_val_list = query_region_count(
            bam_obj=bam_file,
            query_region_list=upstream_region_list,
            norm_method=count_norm_method,
        )

        query_signal_val_list = query_region_count(
            bam_obj=bam_file,
            query_region_list=query_region_list,
            norm_method=count_norm_method,
        )

        down_signal_val_list = query_region_count(
            bam_obj=bam_file,
            query_region_list=downstream_region_list,
            norm_method=count_norm_method,
        )

        merge_sinal_val_list = up_signal_val_list + query_signal_val_list + down_signal_val_list
        if strand == "-":
            merge_sinal_val_list.reverse()
        
        all_info_list.append(region_list + merge_sinal_val_list)


    in_bed_file.close()
    bam_file.close()
    return all_info_list


from ._version import version as __version__
from ._version import author as __author__
from ._version import email as __email__
from ._error import ParsingError
from ._logging import set_logging_level
from ._cli import main as cli_main


from fire import Fire
from .ref_point import main as run_reference_point
from .scale_region import main as run_scale_region

class CLI:
    """Cli interface for python package peaktools.

    - peaktools is a commandline tool and a python package
    """

    def reference_point(
        self,
        input: str,
        input_bam: str,
        output: str,
        output_heatmap: str,
        output_profile: str,
        ref_point: str = "TSS",
        up_extend_length: int = 3000,
        down_extend_length: int = 3000,
        extend_binsize: int = 100,
        count_norm_method: str = "RPKM",
        verbose: str = "ERROR",
        processes: int = 1,
        temp_dir: str = None,
        vmin: int = 0,
        vmax: int = None,
        cmap: str = "Purples",
    ):
        """Calculate matrix and visualize for a reference point.

        Args:
            input (str): Input bed
            input_bam (str): Input bam file
            output (str, optional): Output csv file.
            output_heatmap (str, optional): Output heatmap pdf file.
            output_profile (str, optional): Output profile pdf file.
            up_extend_length (int, optional): upstream region extend length. Defaults to 3000.
            down_extend_length (int, optional): downstream region extend length. Defaults to 3000.
            extend_binsize (int, optional): region extend binsize. Defaults to 100.
            count_norm_method (str, optional): Can be RPKM, Raw or CPM. Defaults to "RPKM".
            verbose (str, optional): The logging level as a string. Defaults to "ERROR".
            processes (int, optional): Number of processes to use. Defaults to 1.
            temp_dir (str, optional): folder to put temp files, use input dir if not defined. Defaults to None.
            vmin (int, optional): Lower boundary of scale bar. Defaults to 0.
            vmax (int, optional): Upper boundary of scale bar. Defaults to 99% quantile.
            cmap (str, optional): Color theme of heatmap. Defaults to "Purples".
        """
        run_reference_point(
            input=input,
            input_bam=input_bam,
            output=output,
            ref_point=ref_point,
            up_extend_length=up_extend_length,
            down_extend_length=down_extend_length,
            extend_binsize=extend_binsize,
            count_norm_method=count_norm_method,
            verbose=verbose,
            processes=processes,
            temp_dir=temp_dir,
            vmin=vmin,
            vmax=vmax,
            cmap=cmap,
            output_heatmap=output_heatmap,
            output_profile=output_profile,
        )

    def scale_region(
        self,
        input: str,
        input_bam: str,
        output: str,
        output_heatmap: str,
        output_profile: str,
        up_extend_length: int = 3000,
        down_extend_length: int = 3000,
        extend_binsize: int = 100,
        split_num: int = 100,
        count_norm_method: str = "RPKM",
        verbose: str = "ERROR",
        processes: int = 1,
        temp_dir: str = None,
        vmin: int = 0,
        vmax: int = None,
        cmap: str = "Purples",
    ):
        """Calculate matrix and visualize for a scaled region.

        Args:
            input (str): Input bed
            input_bam (str): Input bam file
            output (str, optional): Output csv file.
            output_heatmap (str, optional): Output heatmap pdf file.
            output_profile (str, optional): Output profile pdf file.
            up_extend_length (int, optional): upstream region extend length. Defaults to 3000.
            down_extend_length (int, optional): downstream region extend length. Defaults to 3000.
            extend_binsize (int, optional): region extend binsize. Defaults to 100.
            split_num (int, optional): number of bins to split the region. Defaults to 100.
            count_norm_method (str, optional): Can be RPKM, Raw or CPM. Defaults to "RPKM".
            verbose (str, optional): The logging level as a string. Defaults to "ERROR".
            processes (int, optional): Number of processes to use. Defaults to 1.
            temp_dir (str, optional): folder to put temp files, use input dir if not defined. Defaults to None.
            vmin (int, optional): Lower boundary of scale bar. Defaults to 0.
            vmax (int, optional): Upper boundary of scale bar. Defaults to 99% quantile.
            cmap (str, optional): Color theme of heatmap. Defaults to "Purples".
        """
        run_scale_region(
            input=input,
            input_bam=input_bam,
            output=output,
            up_extend_length=up_extend_length,
            down_extend_length=down_extend_length,
            extend_binsize=extend_binsize,
            split_num=split_num,
            count_norm_method=count_norm_method,
            verbose=verbose,
            processes=processes,
            temp_dir=temp_dir,
            vmin=vmin,
            vmax=vmax,
            cmap=cmap,
            output_heatmap=output_heatmap,
            output_profile=output_profile,
        )

def main():
    cli = CLI()
    Fire(cli)

if __name__ == "__main__":
    main()


# python ./query_BAM_with_ref_point.py -h
# usage: query_BAM_with_ref_point.py [-h] -i INPUT -b INPUT_BAM [-o OUTPUT] [--up_extend_length UP_EXTEND_LENGTH]
#                                    [--down_extend_length DOWN_EXTEND_LENGTH] [--extend_binsize EXTEND_BINSIZE]
#                                    [--count_norm_method COUNT_NORM_METHOD] [--verbose VERBOSE]

# Query BAM count info

# options:
#   -h, --help            show this help message and exit
#   -i, --input INPUT     Input bed
#   -b, --input_bam INPUT_BAM
#                         Input bam file
#   -o, --output OUTPUT   Output default=stdout
#   --up_extend_length UP_EXTEND_LENGTH
#                         upstream region extend length, default=3000
#   --down_extend_length DOWN_EXTEND_LENGTH
#                         downstream region extend length, default=3000
#   --extend_binsize EXTEND_BINSIZE
#                         region extend binsize, default=100
#   --count_norm_method COUNT_NORM_METHOD
#                         Can be RPKM, Raw or CPM, default=RPKM
#   --verbose VERBOSE     Larger number means out more log info, can be 0,1,2,3 default=3